package turgen.plugin.kblock;

import turgen.plugin.*;
import turgen.generator.SignalGenerator;
import java.io.*;
import javax.swing.*;
import turgen.Turgen;
import turgen.Utils;
import turgen.dtb.DOS2Binary;
import turgen.playlist.AbstractPlayListItem;
import turgen.config.*;
import turgen.plugin.csturbo.Turbo2000Plugin;
import turgen.tapeimage.TapeImageChunk;

/** Czechoslovak Turbo 2000 - Kilobyte blocks*/
public class KBlockPlugin implements Plugin {
    private KBlockPluginPanel panel;
    
    private final int[] casChunkName = {116, 114, 50, 107};

    private ConfigClass cfcKBlock;
    private IntegerConfigEntry cfgPilotLenA;
    private IntegerConfigEntry cfgPilotLenKBlock;
    private IntegerConfigEntry cfgSilenceAfterLoader;
    private IntegerConfigEntry cfgSilenceAfterHeader;
    private BooleanConfigEntry cfgLoaderNamedAsFile;
    private BooleanConfigEntry cfgPrependLoader;
    private BooleanConfigEntry cfgInvertPolarityOfPulses;

    public static final int BL_NONE = 0;
    public static final int BL_MINITBL = 1;
    public static final int BL_NANOTBL = 2;
    public static final int BL_NANOTBL_UR = 3;
    public static final int BL_NANOTBL_U2 = 4;

    /**
     * Create new instance
     */
    public KBlockPlugin() {
    }

    /** Get plugin name
     * 
     * @return  Plugin name
     */
    public String getName() {
        return "Turbo 2000 - kilobyte blocks";
    }

   /** Convert playlist item
    * 
    * @param _pl Playlist item
    * @return Instructions for generator
    * @throws Exception 
    */
    public int[] convertItem(AbstractPlayListItem _pl) throws Exception {

        KBlockPlayListItem pl;

        pl = (KBlockPlayListItem) _pl;
        RandomAccessFile raf = null;

        try {

            /*Read input file*/
            int[] fileData;
            byte[] b;
            raf = new RandomAccessFile(pl.inFile, "r");
            if (raf.length() > 1024 * 1024) {
                throw new FileTooLargeException(pl.inFile);
            }
            b = new byte[(int) raf.length()];
            raf.readFully(b);

            /*Close input file*/
            raf.close();

            /*Convert to ints*/
            fileData = new int[b.length];
            for (int i = 0; i < b.length; i++) {
                fileData[i] = b[i] < 0 ? b[i] + 256 : b[i];
            }

            /*Check for binary file*/
            int[] initLocations;
            try {
                DOS2Binary dtb = new DOS2Binary(pl.inFile);
                dtb.analyzeFromData(fileData);
                initLocations = dtb.getInitLocations();
            } catch (Exception e) {
                initLocations = new int[0];
            }

            InstructionStream is = new InstructionStream();

            /*If loader is to be prepended, then we prepend it*/
            if (cfgPrependLoader.getAsBoolean() == true) {
                Turbo2000Plugin t2kPlugin = (Turbo2000Plugin) Turgen.pluginManager.getPlugin("turgen.plugin.csturbo.Turbo2000Plugin");
                if (t2kPlugin == null) {
                    throw new PluginException("The Turbo 2000 plugin required to prepend the Universal Turbo loader is not available");
                }
                t2kPlugin.generateUniversalTurboLoader(is);
            }

            /*Prepend binary loader*/
            prependBinaryLoader(is, pl);

            /*Nastaveni*/
            Utils.addSetupInstruction(is,
                    pl.pulses,
                    casChunkName,
                    cfgInvertPolarityOfPulses.getAsBoolean()==false?SignalGenerator.FLAG_POLARITY_10:SignalGenerator.FLAG_POLARITY_01,
                    SignalGenerator.FLAG_ORDER_HL);

            /*Hlavicka*/
            is.add(SignalGenerator.INSTR_PILOTTONE);
            is.add(cfgPilotLenA.getAsInt());
            is.add(SignalGenerator.INSTR_SYNC);
            is.add(SignalGenerator.INSTR_DATA);
            is.add(19);

            int[] header = new int[19];
            header[0] = 0;
            header[1] = 0;

            /*Place blanks to the name*/
            for (int i = 2; i < 18; i++) {
                header[i] = 32;
            }

            /*File name - 16 bytes*/
            int bar = pl.fname.length() < 16 ? pl.fname.length() : 16;
            byte[] nameBytes = pl.fname.substring(0, bar).getBytes();
            
            /*Convert to ints*/
            for (int i = 0; i < nameBytes.length; i++) {
                header[2 + i] = (nameBytes[i] < 0) ? nameBytes[i] + 256 : nameBytes[i];
            }

            /*Header block check sum*/
            int chsum = header[0];
            for (int i = 1; i < 18; i++) {
                chsum = chsum ^ header[i];
            }
            /*Chsum*/
            header[18] = chsum;
            is.add(header);

            /*Safety pulse + safety pause*/
            is.add(SignalGenerator.INSTR_NARROW);
            is.add(SignalGenerator.INSTR_BLOCKSEP);

            /*Another silence after the header*/
            is.add(SignalGenerator.INSTR_LONGSILENCE);
            is.add(cfgSilenceAfterHeader.getAsInt());

            
            /* Position in the data*/
            int ptr = 0;  

            int dl = fileData.length - 1;   /*Index of the last byte*/

            int bcount = 0;

            int startPtr;
            int endPtr;

            /*If not all data converted*/
            while (ptr < fileData.length) {

                startPtr = ptr;

                /*Pilot tone*/
                is.add(SignalGenerator.INSTR_PILOTTONE);
                is.add(cfgPilotLenKBlock.getAsInt());

                /*Sync pulse*/
                is.add(SignalGenerator.INSTR_SYNC);

                /*Block of data*/
                is.add(SignalGenerator.INSTR_DATA);
                is.add(1026);

                /*Block types*/
                /*Full block, at least 1024 bytes to go*/
                if ((dl - ptr + 1) >= 1024) {
                    is.add(255);
                    chsum = 255;
                    bcount = 1024;
                }
                /*Partial or EOF block*/
                else {
                    /*More than 0, partial block*/
                    if ((dl - ptr + 1) > 0) {
                        /*How many pages ?*/
                        is.add(((dl - ptr + 1) / 256) + 251);
                        chsum = ((dl - ptr + 1) / 256) + 251;
                        bcount = dl - ptr + 1;
                    }
                    /*No bytes to go, EOF block*/
                    else {
                        is.add(250);
                        chsum = 250;
                        bcount = 0;
                    }
                }

                /*Block data*/
                for (int i = 0; i < bcount; i++) {
                    chsum = chsum ^ fileData[ptr];
                    is.add(fileData[ptr]);
                    ptr++;

                }
                /*Pad with zeros if needed*/
                for (int i = 0; i < (1023 - bcount); i++) {
                    chsum = chsum ^ 0;
                    is.add(0);
                }

                if (bcount < 1024) {
                    chsum = chsum ^ (bcount % 256);
                    is.add(bcount % 256);
                }

                is.add(chsum);

                endPtr = ptr;

                /*Safety pulse*/
                is.add(SignalGenerator.INSTR_NARROW);

                /*Silence after init or simple block separation*/
                int wait = Utils.calculateInitSilence(startPtr, endPtr - 1, initLocations, pl.silenceList);
                if (wait == 0) {
                    is.add(SignalGenerator.INSTR_BLOCKSEP);
                }
                else {
                    is.add(SignalGenerator.INSTR_LONGSILENCE);
                    is.add(wait);
                }

            }

            /*EOF block*/
            /*Pilot tone*/
            is.add(SignalGenerator.INSTR_PILOTTONE);
            is.add(cfgPilotLenKBlock.getAsInt());
            /*Syncz*/
            is.add(SignalGenerator.INSTR_SYNC);
            is.add(SignalGenerator.INSTR_DATA);
            is.add(1026);
            is.add(250);
            chsum = 250;
            for (int i = 0; i < 1024; i++) {
                chsum = chsum ^ 0;
                is.add(0);
            }
            is.add(chsum);

            /*Safety pulse + safety pause*/
            is.add(SignalGenerator.INSTR_NARROW);
            is.add(SignalGenerator.INSTR_BLOCKSEP);

            is.add(SignalGenerator.INSTR_END);

            return is.getInstructions();

        } catch (Exception e) {

            /*Cleanup*/
            try {
                if (raf!=null) raf.close();
            } catch (IOException xe) {
                
            }
            throw e;

        }

    }

    public AbstractPlayListItem getItemFromPanel() {
        KBlockPlayListItem dpr = panel.getRecord();
        dpr.pluginClass = Turgen.pluginManager.getPluginClassName(this);
        return dpr;
    }

    public void resetPanel() {
        panel.clear();
    }

    public String validateItemOnPanel() {
        return panel.verify();
    }

    public JPanel getItemPanel() {
        return panel;
    }

    public void init() {
        panel = new KBlockPluginPanel();

        /*Initialize configuration class*/
        cfcKBlock = new ConfigClass("KBlock", "Turbo 2000 - kilobyte blocks");

        /*Initialize configuration entries*/
        cfgPilotLenA = new IntegerConfigEntry("pilotLenA", "Header block pilot tone length [pulse]", 3072, 256, 8192);
        cfgPilotLenKBlock = new IntegerConfigEntry("pilotLenBlock", "Data block pilot tone length [pulse]", 2048, 256, 8192);
        cfgSilenceAfterHeader = new IntegerConfigEntry("silenceAfterHeader", "Silence after header [s]", 3, 0, 30);
        cfgSilenceAfterLoader = new IntegerConfigEntry("silenceAfterLoader", "Silence after loader [s]", 3, 0, 30);
        cfgLoaderNamedAsFile = new BooleanConfigEntry("loaderNamedAsFile", "Name loader same as file", false);
        cfgPrependLoader = new BooleanConfigEntry("prependLoader", "Prepend Universal Turbo loader", false);
        cfgInvertPolarityOfPulses = new BooleanConfigEntry("invertPolarity", "Invert polarity of pulses", false);

        /*Add entries to the class*/
        cfcKBlock.addEntry(cfgInvertPolarityOfPulses);
        cfcKBlock.addEntry(cfgPilotLenA);
        cfcKBlock.addEntry(cfgPilotLenKBlock);
        cfcKBlock.addEntry(cfgSilenceAfterHeader);
        cfcKBlock.addEntry(cfgSilenceAfterLoader);
        cfcKBlock.addEntry(cfgLoaderNamedAsFile);
        cfcKBlock.addEntry(cfgPrependLoader);

        /*Register configuration class*/
        Configuration.registerClass(cfcKBlock);
    }

    public void fillPanel(AbstractPlayListItem plr) {
        KBlockPlayListItem dpr = (KBlockPlayListItem) plr;
        panel.fill(dpr);
    }

    public void setInOutProvider(InOutProvider prov) {
        panel.setIOProvider(prov);
    }

    private void prependBinaryLoader(InstructionStream is, KBlockPlayListItem pli) throws Exception {
        
        /*No loader, do nothing*/
        if (pli.loader==BL_NONE) return;
        
        /*Check for turbo 2000 plugin*/
        Turbo2000Plugin t2kPlugin = (Turbo2000Plugin)Turgen.pluginManager.getPlugin("turgen.plugin.csturbo.Turbo2000Plugin");
        if (t2kPlugin == null) {
            throw new PluginException("The Turbo 2000 plugin required to convert the binary loader is not present");
        }
        
        int[] loaderData;
        String fname;
        
        /*Determine which binary loader will be prepended*/
        switch (pli.loader) {
            case BL_MINITBL: {
                if (miniTBLData==null) {
                    miniTBLData = Utils.getLoaderFromResource("/turgen/plugin/kblock/loaders/minitbl.xex");
                }
                loaderData=miniTBLData;
                fname="MiniTBL";
                
                break;
            }
            case BL_NANOTBL: {
                if (nanoTBLData==null) {
                    nanoTBLData = Utils.getLoaderFromResource("/turgen/plugin/kblock/loaders/nanotbl.xex");
                }
                loaderData=nanoTBLData;
                fname="NanoTBL";
                
                break;
            }
            case BL_NANOTBL_UR: {
                if (nanoTBLURData==null) {
                    nanoTBLURData = Utils.getLoaderFromResource("/turgen/plugin/kblock/loaders/nanotbl_ur.xex");
                }
                loaderData=nanoTBLURData;
                fname="NanoTBL";
                
                break;
            }
            case BL_NANOTBL_U2: {
                if (nanoTBLU2Data==null) {
                    nanoTBLU2Data = Utils.getLoaderFromResource("/turgen/plugin/kblock/loaders/nanotbl_u2.xex");
                }
                loaderData=nanoTBLU2Data;
                fname="NanoTBL";
                
                break;
            }
            default: {
                return;
            }
        }
        
        /*Prepend the loader using Turbo 2000 plugin*/
        if (cfgLoaderNamedAsFile.getAsBoolean()==true) {
            fname=pli.fname;
        }
        t2kPlugin.convertMonolithicBinaryUsingData(is, loaderData, fname, true);
        is.add(SignalGenerator.INSTR_LONGSILENCE);
        is.add(cfgSilenceAfterLoader.getAsInt());

    }

    private int[] miniTBLData= null;
    private int[] nanoTBLData = null;
    private int[] nanoTBLURData = null;
    private int[] nanoTBLU2Data = null;

    public String[] getSupportedTapeImageChunks() {
        return new String[0];
    }

    public void addInstructionsForChunk(InstructionStream is, TapeImageChunk chunk) throws Exception {

    }

    public BinaryConvertor[] getBinaryConvertors() {
        BinaryConvertor bcs[] = new BinaryConvertor[5];
        bcs[0] = new KBlockBinaryConvertor(this, "Turbo 2000 KB blocks + MINITBL", 2818, 3489, 1792, 1792 + 1023, BL_MINITBL);
        bcs[1] = new KBlockBinaryConvertor(this, "Turbo 2000 KB blocks + NANOTBL", 2818, 3318, 1792, 1792 + 1023, BL_NANOTBL);
        bcs[2] = new KBlockBinaryConvertor(this, "Turbo 2000 KB blocks + NANOTBL[UR]", 1794, 2349, 49152, 49152 + 1023, BL_NANOTBL_UR);
        bcs[3] = new KBlockBinaryConvertor(this, "Turbo 2000 KB blocks + NANOTBL[U2]", 1794, 2349, 63488, 63488 + 1023, BL_NANOTBL_U2);
        bcs[4] = new KBlockBinaryConvertor(this, "Turbo 2000 KB blocks", 0, 0, 0, 0, BL_NONE);
        return bcs;
    }

    public String getCountry() {
        return "CS";
    }

    public boolean isUsingAlternateFileChooser() {
        return false;
    }

}
